from typing import List, Tuple, Dict, Any
import logging

import pandas as pd
import psqlpandas.postgresql as psql


def drop_table_query(tablename: str) -> str:
    """
    Builds the drop table query.

    Args:
        tablename (str): Table name

    Returns:
        str: drop table query
    """
    return f"DROP TABLE IF EXISTS {tablename}"


def create_table_query(
    tablename: str,
    columns_specs: List[Tuple[str, str]],
    primary_keys: List[str],
) -> str:
    """
    Builds craete table query.

    Args:
        tablename (str): table name
        columns_specs (List[Tuple[str, str]]): spcification of the columns, that is, list of all column names and related types
        primary_keys (List[str]): list of columns to be used as primary keys

    Returns:
        str: the create table query.
    """

    sql = f"CREATE TABLE IF NOT EXISTS {tablename}"
    sql_cols = ", ".join([f"{col} {coltype}" for col, coltype in columns_specs])
    primary_key_sql = f"PRIMARY KEY ({', '.join(primary_keys)})"
    sql = f"{sql} ({sql_cols}, {primary_key_sql})"
    logging.debug(f"Create table query: {sql}")

    return sql


def initialize_table_checks(table_info: Dict[str, Any]) -> bool:
    """
    Checks whether all mandatory keys for the information about tables are present in
    the dictionary passed.

    Args:
        table_info (Dict[str, Any]): dictionary containing the table specifications

    Returns:
        bool: False if any of the mandatory key is missing
    """
    mandatory_keys = ["tablename", "columns", "primary_key", "drop_if_exists"]
    for mk in mandatory_keys:
        if mk not in table_info:
            logging.error(f"Missing mandatory key {mk} in table infos")
            return False
    return True


def initialize_table(
    table_info: Dict[str, Any],
    db_connector: psql.PostgresqlDatabaseConnector,
    data: pd.DataFrame,
) -> bool:
    """
    Initialize a db table.

    Performs the following steps:
    1. drop table if requested (and if it exisits)
    2. create table
    3. upload provided data

    NOTE: if table is not requested to be dropped, input data are inserted in the table
        replacing old values with the same key

    Args:
        table_info (Dict[str, object]): dictionary containing the table specifications
        db_connector (psql.PostgresqlDatabaseConnector): connector to the db
        data (pd.DataFrame): data
    """

    if not initialize_table_checks(table_info):
        return False

    tablename = table_info["tablename"]

    # Drop table
    if table_info["drop_if_exists"]:
        db_connector.execute_sql_query(drop_table_query(tablename))
        logging.debug(f"Dropped table {tablename}")

    # Create table
    db_connector.execute_sql_query(
        create_table_query(
            tablename,
            columns_specs=table_info["columns"].items(),
            primary_keys=table_info["primary_key"]["cols"],
        )
    )

    # Upload data
    db_connector.insert_from_df(
        tablename, data, replace=not table_info["drop_if_exists"]
    )

    return True



from typing import List, Dict, Tuple, Any
import logging

import pandas as pd
import numpy as np
import psycopg2
import psycopg2.extras as pgex
from sqlalchemy import create_engine, text


class PostgresqlDatabaseConnector:
    def __init__(
        self, dbname: str, user: str, host: str, port: str, password: str
    ) -> None:
        self._conn = psycopg2.connect(
            user=user,
            password=password,
            host=host,
            port=port,
            database=dbname,
        )
        self._sqlalchemy_conn = create_engine(
            f"postgresql+psycopg2://{user}:{password}@{host}:{port}/{dbname}"
        )

    def __del__(self) -> None:
        self._conn.close()
        self._sqlalchemy_conn.dispose()

    @staticmethod
    def execute_values(cursor, query, tuples):
        return pgex.execute_values(cursor, query, tuples)

    def read(self, query: str) -> List[Tuple[str]]:
        """
        Reads values from database.

        NOTE: around 3x faster than self.read_df

        Args:
            query (str): select query to be executed

        Returns:
            List[Tuple[str]]: list of rows retrieved
        """
        cur = self._conn.cursor()

        try:
            cur.execute(query)
            data = cur.fetchall()
        except Exception as e:
            logging.error(f"Error while reading query: {query}")
            logging.error(e)
            cur.close()
            raise
        else:
            cur.close()
            return data

    def read_df(
        self,
        query: str,
        parse_dates: List[str] | str | None = None,
        uppercase_colnames: bool = False,
    ) -> pd.DataFrame:
        """
        Reads values from database directly into pandas dataframe

        NOTE: generally slower than self.read. But can be usefult for automatic casting of types (including dates).

        Args:
            query (str): select query to be executed
            parse_dates (List[str] | str | None, optional): columns to be parsed at dates. Defaults to None.
            uppercase_colnames (bool): whether to return uppercase colnames. Defaults to False.

        Returns:
            pd.DataFrame: dataframe of data retrieved
        """
        try:
            data = pd.read_sql_query(
                text(query), con=self._sqlalchemy_conn, parse_dates=parse_dates
            )
            if uppercase_colnames:
                data.columns = data.columns.str.upper()
            return data
        except Exception as e:
            logging.error(f"Error while reading query: {query}")
            logging.error(e)
            raise

    def execute_sql_query(self, query, row=None):
        cur = self._conn.cursor()

        try:
            if row is None:
                cur.execute(query)
            else:
                cur.execute(query, row)
            self._conn.commit()
            cur.close()
            return True
        except Exception as e:
            logging.error(f"Error while executing query: {query}")
            logging.error(e)
            cur.close()
            raise

    def execute_sql_query_values(self, query: str, tuples: Tuple[Any]):
        cur = self._conn.cursor()

        try:
            self.execute_values(cur, query, tuples)
            self._conn.commit()
            cur.close()
            return True
        except Exception as e:
            logging.error(f"Error while executing query: {query}")
            logging.error(e)
            cur.close()
            raise

    def update_db_row(
        self,
        df_dict: Dict[str, object],
        tablename: str,
        sql_where_condition: str | None = None,
    ) -> bool:

        # If you have more than one columns to update
        if len(df_dict.keys()) > 1:
            sql = "UPDATE {} SET ({}) = ({}) ".format(
                tablename,
                ",".join([column for column in df_dict.keys()]),
                ",".join(
                    [
                        f"'{element}'" if element not in [None, []] else "NULL"
                        for element in df_dict.values()
                    ]
                ),
            )
        # If you have only 1 column to modify need to configure a different query
        elif len(df_dict.keys()) == 1:
            sql = f"UPDATE {tablename} SET {list(df_dict.keys())[0]} = {list(df_dict.values())[0]} "
        else:
            raise ValueError(
                "Impossible to generate the update query since data are empty"
            )

        if sql_where_condition is not None:
            sql += sql_where_condition

        return self.execute_sql_query(sql)

    def delete_db_row(
        self,
        tablename: str,
        sql_where_condition: str,
        direct_delete: bool = False,
        update_field: str = "available",
    ) -> bool:
        """
        Delete a database data composing a SQL String with tablename and row id you want to remove

        Args:
            tablename (str): Name of the table
            sql_where_condition (str): SQL Where condition to attach to the query (it's important because you want to delete only a single row).
                Impossible to use the delete without a where setting: for policy and safety purpose.
            direct_delete (bool): If you want to direct delete a row, or just want to update a flag
            update_field (str): Flag field for the deletion (available field column)

        Returns:
            bool: query execution result (True or False)
        """
        # Delete directly the row or bulk of rows
        if direct_delete:
            sql = f"DELETE FROM {tablename} "
        # Just update the flag: available to False
        else:
            sql = f"UPDATE {tablename} SET {update_field} = False "

        # Compose the where condition (it's mandatory to avoid unwanted deletion)
        sql += sql_where_condition

        return self.execute_sql_query(sql)

    def insert_from_df(
        self, tablename: str, df: pd.DataFrame, replace: bool = False
    ) -> bool:
        """
        Inserts data from a pandas dataframe to postgres db.

        Args:
            tablename (str): target tablename
            df (pd.DataFrame): dataframe with data
            replace (bool, optional): whether to replace or not already present data. Defaults to False.

        Returns:
            bool: True if successful without errors
        """

        # duplicate the object because it's required to manipulate some columns
        df_columns = list(set(df.columns) - {"id"})

        # Setting null (and inf) values to None for postgres
        tuples = [
            tuple(x)
            for x in df.astype(object)
            .where(pd.notnull(df) & ~df.isin([np.inf, -np.inf]), None)[df_columns]
            .to_numpy()
        ]

        insertsql = "INSERT INTO {} ({}) VALUES %s".format(
            tablename, ",".join(df_columns)
        )

        # Update if there are data already
        if replace:
            # Query from https://wiki.postgresql.org/wiki/Retrieve_primary_key_columns
            primary_keys_sql = (
                f"SELECT a.attname FROM pg_index i JOIN pg_attribute a ON a.attrelid = i.indrelid AND "
                f"a.attnum = ANY(i.indkey) WHERE i.indrelid = '{tablename}'::regclass AND i.indisprimary;"
            )
            primary_keys = list(self.read_df(primary_keys_sql))
            columns_to_update = [x for x in df_columns if x not in primary_keys]
            insertsql = (
                insertsql
                + " ON CONFLICT ON CONSTRAINT "
                + tablename
                + "_pkey DO UPDATE SET ("
                + ",".join([column for column in columns_to_update])
                + ") = ("
                + ",".join(["EXCLUDED." + column for column in columns_to_update])
                + ")"
            )

        return self.execute_sql_query_values(insertsql, tuples)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['psqlpandas', 'psqlpandas.scripts']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.0.0,<2.0.0',
 'pandas>=2.0.1,<3.0.0',
 'psycopg2>=2.9.5,<3.0.0',
 'sqlalchemy>=2.0.3,<3.0.0']

setup_kwargs = {
    'name': 'psqlpandas',
    'version': '1.2.4',
    'description': 'Utilities to communicate with postgresql database through pandas dataframes.',
    'long_description': '# psqlpandas\nUtilities to communicate with postgresql database through pandas dataframes.\n\n## Installation\n```\npip install psqlpandas\n```\n\n## Usage\nDescribe how to launch and use psqlpandas project.\n',
    'author': 'Mattia Tantardini',
    'author_email': 'mattia.tantardini@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
